{
==============================================================================
 PersistentForm - Release 99/1  27.01.1999
 for Delphi 4, should also run with Delphi 3

 Based on the article "Persistente Komponenten" by Bernhard Angerer in the 
 german developer magazine "Toolbox", issue 1/99

------------------------------------------------------------------------------

 PersistentForm is FREEWARE. Freeware means, that you can use this software 
 without paying anything for it, but it is not public domain! This software 
 is protected through the law of the Federal Republic of Germany, the 
 European Union and other countries.

 (C)1999 by Peter Hellinger Software, All rights are reserved.

 Peter Hellinger Software, Zerzabelshofstrasse 41, D-90480 Nrnberg

 email: mail@helli.de	
 homepage: http://www.helli.de

==============================================================================

 Installation: 
 Put the Unit in the library folder of your delphi compiler. Thats all!

------------------------------------------------------------------------------

 TPersistentForm is a TForm subclass, that remembers the properties of itself 
 and the properties of the components that are put on it. The properties are 
 saved with the destroy event of the form and reloaded with the create event 
 of the form. The informationen is stored likewise in an inifile or in the 
 registry. Output to Inifile is default.

 To use the class simple put the unit PersitentForms in the USES clausel
 of the interface part of your form unit. Replace in the declaration of the
 form class the "class(TForm)" with "class (TPersistentForm)".

 Example:

  type
    TForm1 = CLASS (TForm)  // Without PersistentForm
      ...

  type
    TForm1 = CLASS (TPersistentForm) // With PersistentForm
      ...

 With the OnCreate event of the Form you can select where you want to
 store the form information with the property "direction".

 Example: Saving to inifile

  procedure TForm1.FormCreate(Sender: TObject);
  begin
   direction:= tdInifile; (* not necessary, ini file is default! *)
  end;

 The INI file is created in the directory where the program executable is.
 The name of the inifile is the Application.Title.


 Example: Saving to registry

  procedure TForm1.FormCreate(Sender: TObject);
  begin
   direction:= tdRegistry;
   regkey:= 'Software\Hellinger';
  end;

 The default rootkey is HKEY_CURRENT_USER, this can only changed in the 
 source. Like in the example, you only have to define the subkeys of the 
 registry path, PersistentForms adds the program name (Application.Title)
 automaticly to the path. Each Form that is "persistent" in the application 
 is an own subkey in the application key:

 Software\
          Hellinger\
                    Example\
                            Form1
                            Form2

 Each form key holds the information that is stored.


 These properties are stored from all components:

 'Text'       for TEdits, TCombos etc.
 'Checked'    for TCombos, TMenuItems etc.
 'Position'   for TrackBars
 'InitialDir' for TOpenDialog, TSaveDialog etc
 'FileName'   for TOpenDialog, TSaveDialog etc

 These properties are saved only from TPanel components

 'TPanel.Top', 'TPanel.Left', 'TPanel.Height', 'TPanel.Width'
 
 From the basic form the position and size is always stored

 In the OnCreate event you can easily add more components and properties 
 that should be stored. Simply use the AddFilter methode:

 AddFilter ('Enabled'); // to store the enabled property from ALL components
 AddFilter ('TPanel.Visible'); // to store the visible property of TPanels

 Add properties and components as you like, theres no limit (Ok, in fact
 there IS a limit: How many items can you add to a Stringlist? <grin>)

 
 #############################
 IMPORTANT NOTICE FOR DELPHI4!
 #############################
 
 Inprise changend the way in which forms are created from Delphi 3 to 
 Delphi 4. In Delphi 3, the OnCreate event is generated when in the 
 TPersistentForm.Create procedure the INHERITED methode is executed. In 
 Delphi 4, the OnCreate event first occurs when all Create methodes are 
 completeley executed. That is a problem for PersistentForm, because the 
 settings explained above must be done before the TPersistentForm.Create 
 loads the properties! But Inprise gives us a little help: TForm and is 
 descendants have the property 'OldCreateOrder' to switch back to the
 Delphi 3 createorder.. 

 If you want to make any settings in the OnCreate event of a TPersistentForm
 you MUST set OldCreateOrder at design time of the form to TRUE. Delphi 4
 creates then your form like it does it under Delphi 3.

==============================================================================
}

UNIT PersistentForm;

//------------------------------------------------------------------------------
INTERFACE
//------------------------------------------------------------------------------

USES  Windows, Messages, SysUtils, Classes, Graphics, Controls,
      Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Mask, Menus,
      Registry, Inifiles;

TYPE  TDirection =  (tdRegistry, tdInifile);

TYPE  TPersistentForm = CLASS(TForm)
                         PRIVATE
                          FRegKey:     STRING;
                          RegString:   STRING;
                          FDirection:  TDirection;
                          FFilterList: TStringList;
                          aReg:        TRegistry;
                          aIni:        TInifile;
                          PROCEDURE SaveComps; VIRTUAL;
                          PROCEDURE LoadComps; VIRTUAL;
                          FUNCTION  InFilterList (aCompClassName, aPropName: string): BOOLEAN; VIRTUAL;
                          PROCEDURE SetDefaultFilter; VIRTUAL;
                         PROTECTED
                          PROCEDURE SetDirection (dir: TDirection);
                          PROCEDURE SetRegKey (s: STRING);
                         PUBLIC
                          CONSTRUCTOR Create (AOwner: TComponent); OVERRIDE;
                          DESTRUCTOR Destroy; OVERRIDE;
                          PROCEDURE AddFilter (s: STRING);
                          PROPERTY Direction: TDirection  READ FDirection  WRITE SetDirection;
                          PROPERTY RegKey: STRING  READ  FRegKey  WRITE SetRegKey;
                        END; // TPersistentForm


//------------------------------------------------------------------------------
IMPLEMENTATION
//------------------------------------------------------------------------------

USES TypInfo;


CONST csError =  'TPersistentForm needs an Application.Title.'+#13+'Please assign it with Delphis Project/Options.';
      csTop =    'Top';
      csLeft =   'Left';
      csWidth =  'Width';
      csHeight = 'Height';

CONST tkProperties = [tkFloat, tkInteger, tkChar, tkEnumeration, tkString, tkLString];


FUNCTION GetPropAsString (obj: TObject; info: PPropInfo): STRING;
// Builds from the properties of a component a string
VAR i: LONGINT;
BEGIN
 Result:= '';
 IF info^.PropType^.Kind = tkUnknown THEN EXIT;
 WITH info^ DO BEGIN
  CASE PropType^.Kind OF
   tkEnumeration: IF PropType^.Name = 'Boolean' then BEGIN
                   IF LONGINT (GetOrdProp (obj, info)) = 0 THEN  Result:= 'FALSE'
                                                           ELSE  Result:= 'TRUE';
                   EXIT;
                  END;
   tkInteger:     BEGIN
                   i:= LONGINT (GetOrdProp(Obj,Info));
                   IF (PropType^.Name = 'TColor') OR (PropType^.Name = 'TCursor') THEN BEGIN
                    ColorToIdent (i, Result);
                    EXIT;
                   END;
                   Result:= IntToStr (i);
                  END;
   tkChar:        Result:= CHR (GetOrdProp (obj, info));
   tkString,
   tkLString:     Result:= GetStrProp (obj, info);
  END; // case
 END; // WITH
END; // GetPropAsString


PROCEDURE SetPropFromString (obj: TObject; info: PPropInfo; CONST str: STRING);
// Recreates from a String the properties of a component
VAR i: LONGINT;
BEGIN
 TRY
  WITH Info^ DO BEGIN
   CASE PropType^.Kind OF
    tkEnumeration: IF PropType^.Name = 'Boolean' then BEGIN
                    IF str = 'FALSE' THEN  SetOrdProp (obj, info, 0);
                    IF str = 'TRUE'  THEN  SetOrdProp (obj, info, 1);
                   END;
    tkInteger:     IF str <> '' THEN BEGIN
                    IF PropType^.Name = 'TColor' then BEGIN
                     IF IdentToColor (str, i) THEN  SetOrdProp (obj, info, i)
                                              ELSE  SetOrdProp (obj, info, StrToInt (str));
                     EXIT;
                    END;
                    IF PropType^.Name = 'TCursor' then BEGIN
                     IF IdentToCursor (str, i) THEN  SetOrdProp (obj, info, i)
                                               ELSE  SetOrdProp (obj, info, StrToInt (Str));
                     EXIT;
                    END;
                    SetOrdProp (obj, info, StrToInt (str));
                   END;
    tkChar:        SetOrdProp (obj, info, Ord(str[1]));
    tkString,
    tkLString:     SetStrProp (obj, info, str);
   END; // case
  END; // WITH
 EXCEPT
  // Must gone somthing wrong, huh?
 END;
END; // SetPropFromString


PROCEDURE TPersistentForm.SetDirection (dir: TDirection);
BEGIN
 FDirection:= dir;
END; // SetDirection

PROCEDURE TPersistentForm.SetRegKey (s: STRING);
BEGIN
 FRegKey:= s;
END; // SetRegKey

PROCEDURE TPersistentForm.SaveComps;
// Saves the properties of the components on a form
VAR i, ii, PropAmount: INTEGER;
    PropInfos: PPropList;
    s, str: STRING;
BEGIN
 TRY
  FOR i:= 0 TO ComponentCount-1 DO BEGIN
   str:= '';
   // nil means that it returns all properties
   PropAmount:= GetPropList (Components[i].ClassInfo, tkProperties, NIL);
   IF PropAmount = 0 THEN CONTINUE;
   GetMem (PropInfos, PropAmount * SizeOf (PPropInfo));
   // Get the properties
   TRY
    GetPropList (Components[i].ClassInfo, tkProperties, PropInfos);
    IF FDirection = tdRegistry THEN  aReg.OpenKey(RegString, false); // Write form as key
    FOR ii:= 0 TO PropAmount - 1 DO BEGIN
     IF InFilterList (Components[i].ClassName, PropInfos^[ii]^.Name) THEN BEGIN
      str:= GetPropAsString (Components[i], PropInfos^[ii]);
      IF str <> '' THEN BEGIN
       s:= Components[i].Name + '.' + PropInfos^[ii]^.Name;
       IF FDirection = tdRegistry THEN aReg.WriteString (s, str)
                                  ELSE aIni.WriteString (RegString, s, str);
      END;
     END; // if str
    END; // for ii
    IF FDirection = tdRegistry THEN aReg.CloseKey;
   FINALLY
    FreeMem (PropInfos, PropAmount * SizeOf (PPropInfo));
   END; // try
  END; // for
 FINALLY
 END;
END; // SaveComps


PROCEDURE TPersistentForm.LoadComps;
// Loads the properties of all components of a form
VAR i, ii, PropAmount: INTEGER;
    PropInfos: PPropList;
    s, str: STRING;
BEGIN
 TRY
  FOR i:= 0 TO ComponentCount-1 DO BEGIN
   str:= '';
   // nil means that it returns all properties
   PropAmount:= GetPropList (Components[i].ClassInfo, tkProperties, NIL);
   IF PropAmount = 0 THEN CONTINUE;
   GetMem (PropInfos, PropAmount * SizeOf (PPropInfo));
   // Get the properties
   TRY
    GetPropList (Components[i].ClassInfo, tkProperties, PropInfos);
    IF FDirection = tdRegistry THEN  aReg.OpenKey(RegString, false);
    FOR ii:= 0 TO PropAmount-1 DO BEGIN
     IF InFilterList (Components[i].ClassName, PropInfos^[ii]^.Name) then BEGIN
      s:= Components[i].Name + '.' + PropInfos^[ii]^.Name;
      IF FDirection = tdRegistry THEN str:= aReg.ReadString (s)
                                 ELSE str:= aIni.ReadString (RegString, s, str);
      IF str <> '' then SetPropFromString (Components[i], PropInfos^[ii], str);
     END; // if
    END; // for
    IF FDirection = tdRegistry THEN aReg.CloseKey;
   FINALLY
    FreeMem(PropInfos,PropAmount*SizeOf(PPropInfo));
   END; // try
  END; // for
 FINALLY
 END;
END; // LoadComps


FUNCTION TPersistentForm.InFilterList (aCompClassName, aPropName: STRING): BOOLEAN;
// Returns TRUE, if a property is in the filterlist
VAR i: INTEGER;
BEGIN
 Result:= TRUE;
 IF FFilterList.Find (aPropName, i) then EXIT;
 IF FFilterList.Find (aCompClassName + '.' + aPropName, i) then EXIT;
 Result:= false;
END; // InfilterList

PROCEDURE TPersistentForm.SetDefaultFilter;
// Initialize the filterlist
BEGIN
 WITH FFilterList DO BEGIN
  // properties that should be saved from ALL components
  Add ('Text');          // for TEdits, TCombos etc.
  Add ('Checked');       // for TCombos, TMenuItems etc.
  Add ('Position');      // for TRackBars
  Add ('InitialDir');    // for TOpENDialog, TSaveDialog etc
  Add ('FileName');      // for TOpENDialog, TSaveDialog etc

  // properties that should be saved only from SPECIFIED components
  Add ('TPanel.Top');
  Add ('TPanel.Left');
  Add ('TPanel.Height');
  Add ('TPanel.Width');
  Sort;
 END;
END; // SetDefaultFilter

PROCEDURE TPersistentForm.AddFilter (s: STRING);
// Adds a filter to the filterlist
BEGIN
 FFilterList.Add (s);
END; // AddFilter

DESTRUCTOR TPersistentForm.Destroy;
// saves the properties and destroys the form
BEGIN
 IF FRegKey <> '' then BEGIN
  IF FDirection = tdRegistry THEN BEGIN
   aReg:= TRegistry.Create;
   aReg.RootKey := HKEY_CURRENT_USER;
   RegString:= FRegKey + Application.Title + '\' + Self.Name;
   IF aReg.OpenKey(RegString, TRUE) THEN BEGIN
    // Save position and size of the form
    aReg.WriteInteger (csTop, Self.Top);
    aReg.WriteInteger (csLeft, Self.Left);
    aReg.WriteInteger (csWidth, Self.Width);
    aReg.WriteInteger (csHeight, Self.Height);
    SaveComps; // Save the properties of the components
   END;
   aReg.Free;
  END ELSE BEGIN
   RegString:= Self.Name;
   aIni:= TInifile.Create (FRegKey);
   // Save position and size of the form
   aIni.WriteInteger (Self.Name, csTop, Self.Top);
   aIni.WriteInteger (Self.Name, csLeft, Self.Left);
   aIni.WriteInteger (Self.Name, csWidth, Self.Width);
   aIni.WriteInteger (Self.Name, csHeight, Self.Height);
   SaveComps; // Save the properties of the components
   aIni.Free;
  END;
 END;
 INHERITED Destroy;
END; // Destroy


CONSTRUCTOR TPersistentForm.Create (AOwner: TComponent);
// creates the form and restores the properties
VAR t, l, w, h: INTEGER;
BEGIN
 FRegKey:= '';
 RegString:= '';
 FFilterList:= TStringList.Create;
 FFilterList.Sorted:= TRUE; // Always sort the list
 FDirection:= tdInifile;
 SetDefaultFilter;

// IMPORTANT NOTICE FOR DELPHI4!
// =============================
// Inprise changend the way in which forms are created from Delphi 3 to 
// Delphi 4. In Delphi 3, the OnCreate event is generated when in the 
// TPersistentForm.Create the INHERITED methode is executed. In Delphi 4, 
// the OnCreate event first occurs when all Create methodes are completeley 
// executed. That is a problem for PersistentForm, because the settings
// explained above must be done before the TPersistentForm.Create loads
// the properties! But Inprise gives us a little help: TForm and is 
// descendants have the property 'OldCreateOrder'. 
//
// If you want to make any settings in the OnCreate event of a TPersistentForm
// you MUST set OldCreateOrder at design time of the form to TRUE. Delphi 4
// creates then your form like it does it under Delphi 3.

 INHERITED; // Generate the OnCreate event

 IF FDirection = tdRegistry THEN BEGIN
  IF FRegKey = '' THEN  FRegKey:= 'Software\' + Application.Title;
  IF FRegKey = '' THEN BEGIN  ShowMessage (csError);  EXIT;  END;
  RegString:= FRegKey + Application.Title + '\' + Self.Name;
  aReg:= TRegistry.Create;
  aReg.RootKey:= HKEY_CURRENT_USER;
  IF aReg.OpenKey (RegString, FALSE) THEN BEGIN
   TRY  t:= aReg.ReadInteger (csTop);    EXCEPT  t:= 0;  END;
   TRY  l:= aReg.ReadInteger (csLeft);   EXCEPT  l:= 0;  END;
   TRY  w:= aReg.ReadInteger (csWidth);  EXCEPT  w:= 0;  END;
   TRY  h:= aReg.ReadInteger (csHeight); EXCEPT  h:= 0;  END;
   LoadComps; // Restore the properties of the components
  END;
  aReg.Destroy;
 END ELSE BEGIN
  IF FRegKey = '' then FRegKey:= ExtractFilePath (Application.ExeName) + Application.Title + '.ini';
  IF FRegKey = '' then BEGIN  ShowMessage(csError);  EXIT;  END;
  RegString:= Self.Name;
  aIni:= TInifile.Create (FRegKey);
  t:= aIni.ReadInteger (Self.Name, csTop, 0);
  l:= aini.ReadInteger (Self.Name, csLeft, 0);
  w:= aIni.ReadInteger (Self.Name, csWidth, 0);
  h:= aIni.ReadInteger (Self.Name, csHeight, 0);
  LoadComps; // Restore the properties of the components
  aIni.Free;
 END;
 // Restore position and size of the form
 IF t <> 0 THEN  Self.Top:= t;
 IF l <> 0 THEN  Self.Left:= l;
 IF w <> 0 THEN  Self.Width:= w;
 IF h <> 0 THEN  Self.Height:= h;
END; // Create

END. // PersitentForms
//==============================================================================
